@@ -51,6 +51,7 @@ class Alignment {
         vector<double> get_frequencies();
         vector<string> get_names();
         size_t get_number_of_sequences();
+        vector<string> get_sites();
         size_t get_number_of_sites();
         size_t get_number_of_free_parameters();
         string get_substitution_model();
@@ -90,6 +91,7 @@ class Alignment {
 
         // Bootstrap
         vector<pair<string, string>> get_bootstrapped_sequences();
+        void chkdst();
 
     private :
         string _get_datatype();
@@ -114,6 +116,7 @@ class Alignment {
         shared_ptr<NNIHomogeneousTreeLikelihood> likelihood = nullptr;
         shared_ptr<HomogeneousSequenceSimulator> simulator = nullptr;
         string _name;
+        string _computeTree(DistanceMatrix dists, DistanceMatrix vars) throw (Exception);
 };
 
 #endif /* _ALIGNMENT_H_ */
